import logging
import os
import sys


class StdOutHandler(logging.Handler):
    """
    Because we re-bind sys.stdout/stderr dynamically to redirect
    print statements to Girder job logs, we must also dynamically
    refer to sys.stdout dynamically in our custom log handler.
    """
    def emit(self, record):
        sys.stdout.write(self.format(record) + os.linesep)


def setupLogger(config):
    level = getattr(logging, config.get('logging', 'level').upper())
    logger = logging.getLogger('girder_worker')
    logger.setLevel(level)

    handler = StdOutHandler()
    formatter = logging.Formatter(config.get('logging', 'format'))
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger



from . import StreamReader


class GirderFileStreamReader(StreamReader):
    """
    Stream a file from Girder.
    """
    def __init__(self, client, file_id):
        """
        :param client: The GirderClient instance to use.
        :type client: girder_client.GirderClient
        """
        super().__init__()
        self._client = client
        self._file_id = file_id
        self._iter = None

    def read(self, buf_len):
        """
        Implementation note: due to a constraint of the requests library, the
        buf_len that is used the first time this method is called will cause
        all future requests to ``read`` to have the same ``buf_len`` even if
        a different ``buf_len`` is passed in on subsequent requests.
        """
        if self._iter is None:  # lazy load response body iterator
            self._iter = self._client.downloadFileAsIterator(self._file_id, buf_len)

        try:
            return next(self._iter)
        except StopIteration:
            return b''



import os
import errno
import stat
import abc
import urllib
import ssl
import sys
from http import client as httplib


class FDStreamConnector(metaclass=abc.ABCMeta):
    """
    FDStreamConnector is an abstract base class used to connect a read(input) and write(output)
    stream.
    """

    def __init__(self, input, output):
        self.input = input
        self.output = output

    @abc.abstractmethod
    def open(self):
        """
        Open the stream connector, delegated to implementation.
        """

    @abc.abstractmethod
    def close(self):
        """
        Close the stream connector, delegated to implementation.
        """

    @abc.abstractmethod
    def fileno(self):
        """
        This method allows an instance of this class to used in the select call.

        :returns: The file descriptor that should be used to determine if the connector
                 has data to process.
        """


class FDWriteStreamConnector(FDStreamConnector):
    """
    FDWriteStreamConnector can be used to connect a read and write stream. The write
    side of the connection will be used in the select loop to trigger write i.e
    the file descriptor from the write stream will be used in the select call.

    This is typically used to stream data to a named pipe that will be read inside
    a container.
    """

    def fileno(self):
        """
        This method allows an instance of this class to used in the select call.

        :returns: The file descriptor for write(output) side of the connection.
        """
        return self.output.fileno()

    def write(self, n=65536):
        """
        Called when it is detected the output side of this connector is ready
        to write. Reads (potentially blocks) at most n bytes and writes them to
        the output ends of this connector. If no bytes could be read, the connector
        is closed.

        :param n The maximum number of bytes to write.
        :type n int
        :returns: The actual number of bytes written.
        """
        buf = self.input.read(n)

        if buf:
            return self.output.write(buf)
        else:
            self.close()

        return 0

    def open(self):
        """
        Calls open on the output side of this connector.
        """
        self.output.open()

    def close(self):
        """
        Closes the output side of this connector, followed by the input side.
        """
        self.output.close()
        if hasattr(self.input, 'close'):
            self.input.close()


class FDReadStreamConnector(FDStreamConnector):
    """
    FDReadStreamConnector can be used to connect a read and write stream. The read
    side of the connection will be used in the select loop to trigger write i.e
    the file descriptor from the read stream will be used in the select call.

    This is typically used to stream data from a named pipe that is being written
    to inside a container.
    """

    def fileno(self):
        """
        This method allows an instance of this class to used in the select call.

        :returns: The file descriptor for read(input) side of the connection.
        """

        return self.input.fileno()

    def read(self, n=65536):
        """
        Called when it is detected the input side of this connector is ready
        to read. Reads at most n bytes and writes them to the output ends of
        this connector. If no bytes could be read, the connector is closed.

        :param n The maximum number of bytes to read.
        :type n int
        :returns: The actual number of bytes read.
        """
        buf = self.input.read(n)

        if buf:
            self.output.write(buf)
            # TODO PushAdapter/Writers should return number of bytes actually
            # written.
            return len(buf)
        else:
            self.close()

        return 0

    def open(self):
        """
        Calls open on the input side of this connector.
        """
        self.input.open()

    def close(self):
        """
        Closes the output side of this connector, followed by the input side.
        """
        self.input.close()
        self.output.close()


class StreamReader(metaclass=abc.ABCMeta):
    """
    This represents the interface that must be implemented by a stream reader.
    """

    @abc.abstractmethod
    def read(self, buf_len):
        """
        Stream readers must implement this method, which is responsible for
        reading up to ``buf_len`` bytes from the stream. For now, this is
        expected to be a blocking read, and should return an empty string to
        indicate the end of the stream.
        """

    def close(self):
        """
        Close the input stream. Called after the last data is read.
        """
        pass


class StreamWriter(metaclass=abc.ABCMeta):
    """
    This represents the interface that must be implemented by a stream writer.
    """

    @abc.abstractmethod
    def write(self, buf):
        """
        Write a chunk of data to the output stream.
        """

    def close(self):
        """
        Close the output stream. Called after the last data is sent.
        """
        pass


class FileDescriptorReader(StreamReader):
    """
    Reader to read from a file descriptor.
    """
    def __init__(self, fd):
        self._fd = fd

    def fileno(self):
        return self._fd

    def read(self, n):
        return os.read(self.fileno(), n)

    def close(self):
        os.close(self.fileno())


class FileDescriptorWriter(StreamWriter):
    """
    Writer to write to a file descriptor.
    """
    def __init__(self, fd):
        self._fd = fd

    def fileno(self):
        return self._fd

    def write(self, buf):
        return os.write(self.fileno(), buf)

    def close(self):
        os.close(self.fileno())


class StdStreamWriter(StreamWriter):
    """
    Writer for write to stdout and stderr.
    """
    def __init__(self, stream):
        self._stream = stream

    def write(self, buf):
        return self._stream.write(buf)

    def close(self):
        # Don't close std streams!
        self._stream.flush()


class NamedPipe:
    """
    A named pipe.
    """
    def __init__(self, path):
        self.path = path
        self._fd = None
        os.mkfifo(self.path)

    def open(self, flags):

        if self._fd is None:
            if not os.path.exists(self.path):
                raise Exception('Input pipe does not exist: %s' % self._path)
            if not stat.S_ISFIFO(os.stat(self.path).st_mode):
                raise Exception('Input pipe must be a fifo object: %s' % self._path)

            try:
                self._fd = os.open(self.path, flags)
            except OSError as e:
                if e.errno != errno.ENXIO:
                    raise e

    def fileno(self):
        return self._fd


class NamedPipeReader(FileDescriptorReader):
    """
    Reader to read from a named pipe.
    """
    def __init__(self, pipe, container_path=None):
        super().__init__(None)
        self._pipe = pipe
        self._container_path = container_path

    def open(self):
        self._pipe.open(os.O_RDONLY | os.O_NONBLOCK)

    def path(self):
        """
        The argument to pass to the docker container's entrypoint. In this case
        the path to the named pipe.
        """
        return self._container_path

    def fileno(self):
        return self._pipe.fileno()


class NamedPipeWriter(FileDescriptorWriter):
    """
    Write to write to a named pipe.
    """
    def __init__(self, pipe, container_path=None):
        super().__init__(None)
        self._pipe = pipe
        self._container_path = container_path

    def open(self):
        self._pipe.open(os.O_WRONLY | os.O_NONBLOCK)

    def path(self):
        """
        The argument to pass to the docker container's entrypoint. In this case
        the path to the named pipe.
        """
        return self._container_path

    def fileno(self):
        return self._pipe.fileno()


class ChunkedTransferEncodingStreamWriter(StreamWriter):
    def __init__(self, url, headers={}):
        self._url = url

        """
        Uses HTTP chunked transfer-encoding to stream a request body to a
        server. Unfortunately requests does not support hooking into this logic
        easily, so we use the lower-level httplib module.
        """
        self._closed = False

        parts = urllib.parse.urlparse(self._url)
        if parts.scheme == 'https':
            ssl_context = ssl.create_default_context()
            conn = httplib.HTTPSConnection(parts.netloc, context=ssl_context)
        else:
            conn = httplib.HTTPConnection(parts.netloc)

        try:
            url = parts.path
            if parts.query is not None:
                url = '%s?%s' % (url, parts.query)
            conn.putrequest('POST',
                            url, skip_accept_encoding=True)

            for header, value in headers.items():
                conn.putheader(header, value)

            conn.putheader('Transfer-Encoding', 'chunked')

            conn.endheaders()  # This actually flushes the headers to the server
        except Exception:
            sys.stderr.write('HTTP connection to "%s" failed.\n' % self._url)
            conn.close()
            raise

        self.conn = conn

    def write(self, buf):
        """
        Write a chunk of data to the output stream in accordance with the
        chunked transfer encoding protocol.
        """
        try:
            self.conn.send(hex(len(buf))[2:].encode('utf-8'))
            self.conn.send(b'\r\n')
            self.conn.send(buf)
            self.conn.send(b'\r\n')
        except Exception:
            resp = self.conn.getresponse()
            sys.stderr.write(
                'Exception while sending HTTP chunk to %s, status was %s, '
                'message was:\n%s\n' % (self._url, resp.status,
                                        resp.read()))
            self.conn.close()
            self._closed = True
            raise

    def close(self):
        """
        Close the output stream. Called after the last data is sent.
        """
        if self._closed:
            return

        try:
            self.conn.send(b'0\r\n\r\n')
            resp = self.conn.getresponse()
            if resp.status >= 300 and resp.status < 400:
                raise Exception('Redirects are not supported for streaming '
                                'requests at this time. %d to Location: %s' % (
                                    resp.status, resp.getheader('Location')))
            if resp.status >= 400:
                raise Exception(
                    'HTTP stream output to %s failed with status %d. Response '
                    'was: %s' % (
                        self.output_spec['url'], resp.status, resp.read()))
        finally:
            self.conn.close()



import abc
import os
import stat
import sys
import tempfile
import uuid

from girder_worker_utils.transform import Transform

TEMP_VOLUME_MOUNT_PREFIX = '/mnt/girder_worker'


def _maybe_transform(obj, *args, **kwargs):
    if hasattr(obj, 'transform') and callable(obj.transform):
        return obj.transform(*args, **kwargs)

    return obj


class HostStdOut(Transform):
    """
    Represents the standard output stream on the host machine. Can be used with
    :py:class:`girder_worker.docker.transforms.Connect` to write text to stdout.
    """
    def transform(self, **kwargs):
        from girder_worker.docker.io import (
            StdStreamWriter
        )
        return StdStreamWriter(sys.stdout)


class HostStdErr(Transform):
    """
    Represents the standard error stream on the host machine. Can be used with
    :py:class:`girder_worker.docker.transforms.Connect` to write text to stderr.
    """
    def transform(self, **kwargs):
        from girder_worker.docker.io import (
            StdStreamWriter
        )
        return StdStreamWriter(sys.stderr)


class ContainerStdOut(Transform):
    """
    Represents the standard output stream of the container. Can be used with
    :py:class:`girder_worker.docker.transforms.Connect` to redirect the containers
    standard output to another stream.
    """
    def transform(self, **kwargs):
        return self

    def open(self):
        # noop
        pass


class ContainerStdErr(Transform):
    """
    Represents the standard error stream of the container. Can be used with
    :py:class:`girder_worker.docker.transforms.Connect` to redirect the containers
    standard error to another stream.
    """
    def transform(self, **kwargs):
        return self

    def open(self):
        # noop
        pass


class BindMountVolume(Transform):
    """
    A volume that will be bind mounted into a docker container.

    :param host_path: The path on the host machine.
    :type host_path: str
    :param container_path: The path in the container this volume will be mounted
        at.
    :type container_path: str
    :param mode: The mounting mode
    :type mode: str
    """
    def __init__(self, host_path, container_path, mode='rw'):
        self._host_path = host_path
        self._container_path = container_path
        self.mode = mode

    def _repr_json_(self):
        return {
            self.host_path: {
                'bind': self.container_path,
                'mode': self.mode
            }
        }

    def transform(self, **kwargs):
        return self.container_path

    @property
    def container_path(self):
        return self._container_path

    @property
    def host_path(self):
        return self._host_path


class _TemporaryVolumeMetaClass(abc.ABCMeta):
    @property
    def default(cls):
        """
        This returns the default temporary volume that is always mounted into the container.
        """
        return _DefaultTemporaryVolume()


class _TemporaryVolumeBase(BindMountVolume):
    def __init__(self, *arg, **kwargs):
        super().__init__(*arg, **kwargs)
        self._transformed = False

    def _make_paths(self, host_dir=None, mode=0o755):
        if host_dir is not None and not os.path.exists(host_dir):
            os.makedirs(host_dir)
            # Sometimes we need to explicitly set the mode of the
            # directory to 0o777 (e.g. when running the integration
            # tests). To do this explicitly (without the user's umask
            # getting in the way) we must make a separate call to
            # os.chmod.
            os.chmod(host_dir, mode)
        self._host_path = tempfile.mkdtemp(dir=host_dir)
        try:
            # this is permissive as a worker may be running as a different user
            # and could otherwise not read or create files.
            os.chmod(self._host_path, 0o777 | stat.S_ISGID)
        except Exception:
            pass
        self._container_path = os.path.join(TEMP_VOLUME_MOUNT_PREFIX, uuid.uuid4().hex)


class TemporaryVolume(_TemporaryVolumeBase, metaclass=_TemporaryVolumeMetaClass):
    """
    This is a class used to represent a temporary directory on the host that will
    be mounted into a docker container. girder_worker will automatically attach a default
    temporary volume. This can be reference using `TemporaryVolume.default` class attribute.
    A temporary volume can also be create in a particular host directory by providing the
    `host_dir` param.

    :param host_dir: The root directory on the host to use when creating the
        the temporary host path.
    :type host_dir: str
    :param mode: The default mode applied to the temporary volume if it does
        not already exist.
    :type mode: int
    """
    # Note that this mode is explicitly set with os.chmod. What you
    # set, is what you get - no os.makedirs umask shenanigans.
    def __init__(self, host_dir=None, mode=0o777):
        super().__init__(None, None)
        self.host_dir = host_dir
        self._mode = mode
        self._instance = None
        self._transformed = False

    def transform(self, **kwargs):
        if not self._transformed:
            self._transformed = True
            self._make_paths(self.host_dir, mode=self._mode)

        return super().transform(**kwargs)


class _DefaultTemporaryVolume(TemporaryVolume):
    """
    Place holder who delegates implementation to instance provide by transform(...) method
    An instance of the class is returned each time `TemporaryVolume.default` is accessed.
    When the docker_run task is executed the transform(...) method is call with an instance
    containing information about the actual default temporary volume associated with the
    task. The place holder then delegates all functionality to this instance.
    """
    def transform(self, _default_temp_volume=None, **kwargs):
        self._instance = _default_temp_volume
        self._transformed = True

        return self._instance.transform(**kwargs)

    @property
    def container_path(self):
        return self._instance.container_path

    @property
    def host_path(self):
        return self._instance.host_path


class NamedPipeBase(Transform):
    def __init__(self, name, container_path=None, host_path=None, volume=TemporaryVolume.default):
        super().__init__()
        self._container_path = None
        self._host_path = None
        self._volume = None

        if container_path is not None and host_path is not None:
            self._container_path = container_path
            self._host_path = host_path
        else:
            self._volume = volume

        self.name = name

    def transform(self, **kwargs):
        if self._volume is not None:
            self._volume.transform(**kwargs)

    @property
    def container_path(self):
        """
        The path within the docker container.
        """
        if self._container_path is not None:
            return os.path.join(self._container_path, self.name)
        else:
            return os.path.join(self._volume.container_path, self.name)

    @property
    def host_path(self):
        """
        The path on the host machine
        """
        if self._host_path is not None:
            return os.path.join(self._host_path, self.name)
        else:
            return os.path.join(self._volume.host_path, self.name)

    def cleanup(self, **kwargs):
        os.remove(self.host_path)


class NamedInputPipe(NamedPipeBase):
    """
    A named pipe that can be open for read within a docker container.
    i.e. To stream data into a container.

    :param name: The name of the pipe.
    :type name: str
    :param container_path: The path in the container.
    :type container_path: str
    :param host_path: The path on the host machine.
    :type host_path: str
    :param volume: Alternatively a :py:class:`girder_worker.docker.transforms.BindMountVolume`
        instance can be provided. In which case the container_path and host_paths from
        the volume will be used when creating the pipe. The default location is
        :py:obj:`girder_worker.docker.transforms.TemporaryVolume.default`
    """
    def __init__(self, name,  container_path=None, host_path=None, volume=TemporaryVolume.default):
        super().__init__(name, container_path, host_path, volume)

    def transform(self, **kwargs):
        from girder_worker.docker.io import (
            NamedPipe,
            NamedPipeWriter
        )
        super().transform(**kwargs)

        pipe = NamedPipe(self.host_path)
        return NamedPipeWriter(pipe, self.container_path)


class NamedOutputPipe(NamedPipeBase):
    """
    A named pipe that can be opened for write within a docker container.
    i.e. To stream data out of a container.

    :param name: The name of the pipe.
    :type name: str
    :param container_path: The path in the container.
    :type container_path: str
    :param host_path: The path on the host machine.
    :type host_path: str
    :param volume: Alternatively a :py:class:`girder_worker.docker.transforms.BindMountVolume`
        instance can be provided. In which can the container_path and host_paths from
        the volume will be use when creating the pipe. The default location is
        :py:attr:`girder_worker.docker.transforms.TemporaryVolume.default`
    """
    def __init__(self, name, container_path=None, host_path=None, volume=TemporaryVolume.default):
        super().__init__(name, container_path, host_path, volume)

    def transform(self, **kwargs):
        from girder_worker.docker.io import (
            NamedPipe,
            NamedPipeReader
        )
        super().transform(**kwargs)

        pipe = NamedPipe(self.host_path)
        return NamedPipeReader(pipe, self.container_path)


class VolumePath(Transform):
    """
    A path on a docker volume. Must be a path relative to the root of the volume.

    :param filename: The file name.
    :type name: str
    :param volume: The volume this file lived on. If no volume is provided then
        the file will be on
        :py:attr:`girder_worker.docker.transforms.TemporaryVolume.default`
    :type volume: :py:class:`girder_worker.docker.transforms.BindMountVolume`
    """
    def __init__(self, filename, volume=TemporaryVolume.default):
        if os.path.isabs(filename):
            raise Exception('VolumePath paths must be relative to a volume (%s).' % filename)

        self.filename = filename
        self._volume = volume

    def transform(self, *pargs, **kwargs):
        self._volume.transform(**kwargs)
        # If we are being called with arguments, then this is the execution of
        # girder_result_hooks, so return the host_path
        if len(pargs) > 0:
            return os.path.join(self._volume.host_path, self.filename)
        else:
            return os.path.join(self._volume.container_path, self.filename)

    def _repr_model_(self):
        return '<%s.%s: "%s">' % (self.__module__, self.__class__.__name__, self.filename)


class Connect(Transform):
    """
    This utility class represents the connection between a
    :py:class:`girder_worker.docker.transforms.NamedOutputPipe` or
    :py:class:`girder_worker.docker.transforms.NamedInputPipe` and one of the other streaming
    transforms. Girder Worker will stream the data to or from the named pipe.

    :param input: The input side of the connection
    :type input: :py:class:`girder_worker.docker.transforms.NamedOutputPipe` or
        :py:class:`girder_worker.docker.transforms.girder.GirderFileIdToStream`
    :param output: The output side of the connection
    :type output: :py:class:`girder_worker.docker.transforms.NamedInputPipe` or
        :py:class:`girder_worker.docker.transforms.ChunkedTransferEncodingStream` or
        :py:class:`girder_worker.docker.transforms.HostStdOut` or
        :py:class:`girder_worker.docker.transforms.HostStdErr`
    """
    def __init__(self, input, output):
        super().__init__()
        self._input = input
        self._output = output

    def transform(self, **kwargs):
        from girder_worker.docker.io import (
            FDWriteStreamConnector,
            FDReadStreamConnector,
        )
        input = _maybe_transform(self._input, **kwargs)
        output = _maybe_transform(self._output, **kwargs)
        if isinstance(self._output, NamedInputPipe):
            return FDWriteStreamConnector(input, output)
        elif isinstance(self._input, NamedOutputPipe):
            return FDReadStreamConnector(input, output)
        else:
            raise TypeError('A NamedInputPipe or NamedOutputPipe must be provided.')

    def _repr_model_(self):
        """
        The method is called before save the argument in the job model.
        """
        return str(self)


class ChunkedTransferEncodingStream(Transform):
    """
    A stream transform that allows data to be streamed using HTTP Chunked Transfer Encoding
    to a server.

    :param url: Destination URL for the stream.
    :type url: str
    :param headers: HTTP headers to send.
    :type header: dict
    """
    def __init__(self, url, headers={}, **kwargs):
        self.url = url
        self.headers = headers

    def transform(self, **kwargs):
        from girder_worker.docker.io import (
            ChunkedTransferEncodingStreamWriter
        )

        return ChunkedTransferEncodingStreamWriter(self.url, self.headers)
